import unittest
from unittest.mock import MagicMock, patch

from lifeguard import NORMAL, PROBLEM
from lifeguard.notifications import (
    NotificationBase,
    NotificationStatus,
)
from lifeguard.actions.notifications import (
    notify_in_single_message,
    notify_in_thread,
)

mock_implementation = MagicMock(name="mock_implementation")
mock_logger = MagicMock(name="logger")


class MockNotificationBase(NotificationBase):
    def __init__(self):
        self.mock_implementation = None

    def send_single_message(self, content, settings):
        super().send_single_message(content, settings)
        self.mock_implementation.send_single_message(content, settings)

    def init_thread(self, content, settings):
        super().init_thread(content, settings)
        self.mock_implementation.init_thread(content, settings)

    def update_thread(self, thread_id, content, settings):
        super().update_thread(thread_id, content, settings)
        self.mock_implementation.update_thread(thread_id, content, settings)

    def close_thread(self, thread_id, content, settings):
        super().close_thread(thread_id, content, settings)
        self.mock_implementation.close_thread(thread_id, content, settings)


mock_notification_instance = MockNotificationBase()
mock_notification_instance.mock_implementation = mock_implementation


NOTIFICATION_METHODS = [mock_notification_instance]


class TestActionNotification(unittest.TestCase):
    def setUp(self):
        NOTIFICATION_METHODS[0].mock_implementation.reset_mock()
        self.mock_validation_response = MagicMock(name="mock_validation_response")
        self.mock_validation_response.settings = None
        self.mock_settings = {}

    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", [])
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_not_send_notify_in_single_message(self):
        notify_in_single_message(self.mock_validation_response, self.mock_settings)
        mock_implementation.send_single_message.assert_not_called()

    @patch("lifeguard.actions.notifications.HistoryRepository")
    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_send_notify_in_single_message(self, mock_history_repository):
        history_repository_instance = MagicMock(name="history_repository")
        mock_history_repository.return_value = history_repository_instance

        self.mock_validation_response.validation_name = "validation_name"
        self.mock_validation_response.settings = {"notification": {"notify": True}}
        self.mock_validation_response.details = {"status": "PROBLEM"}
        self.mock_validation_response.status = PROBLEM

        notify_in_single_message(
            self.mock_validation_response, {"notification": {"add_to_history": True}}
        )

        mock_implementation.send_single_message.assert_called_with(
            '{"status": "PROBLEM"}', {"notification": {"add_to_history": True}}
        )
        notification_occurrence = (
            history_repository_instance.append_notification.call_args[0][0]
        )

        self.assertEqual(notification_occurrence.notification_type, "single")
        self.assertEqual(notification_occurrence.validation_name, "validation_name")
        self.assertEqual(notification_occurrence.details, {"status": "PROBLEM"})
        self.assertEqual(notification_occurrence.status, "PROBLEM")

    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_send_notify_in_single_message_with_template(self):
        self.mock_validation_response.settings = {
            "notification": {
                "notify": True,
                "data": {"user": "a user", "body": "message body"},
            }
        }
        self.mock_validation_response.details = {"status": "PROBLEM"}

        self.mock_settings["notification"] = {
            "template": """Hello {{user}},
{{body}}
"""
        }

        notify_in_single_message(self.mock_validation_response, self.mock_settings)

        mock_implementation.send_single_message.assert_called_with(
            "Hello a user,\nmessage body",
            {"notification": {"template": "Hello {{user}},\n{{body}}\n"}},
        )

    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_send_multiple_notify_in_single_message_with_template(self):
        self.mock_validation_response.settings = {
            "notification": {
                "notify": True,
                "data": [
                    {"user": "first user", "body": "message body"},
                    {"user": "second user", "body": "message body"},
                ],
            }
        }
        self.mock_validation_response.details = {"status": "PROBLEM"}

        self.mock_settings["notification"] = {
            "template": """Hello {{user}},
{{body}}
"""
        }

        notify_in_single_message(self.mock_validation_response, self.mock_settings)

        mock_implementation.send_single_message.assert_called_with(
            ["Hello first user,\nmessage body", "Hello second user,\nmessage body"],
            {"notification": {"template": "Hello {{user}},\n{{body}}\n"}},
        )

    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_not_send_notify_in_single_if_is_disabled(self):
        self.mock_validation_response.settings = {
            "notification": {
                "notify": True,
                "data": {"user": "a user", "body": "message body"},
            }
        }
        self.mock_validation_response.details = {"status": "PROBLEM"}

        self.mock_settings["notification"] = {
            "disabled": ["not implemented"],
            "template": """Hello {{user}},
{{body}}
""",
        }

        notify_in_single_message(self.mock_validation_response, self.mock_settings)

        mock_implementation.send_single_message.assert_not_called()

    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    @patch("lifeguard.actions.notifications.NotificationRepository")
    def test_should_send_notify_in_thread_message_when_error_using_traceback(
        self, mock_notification_repository
    ):
        notification_repository_instance = MagicMock(name="repository")
        mock_notification_repository.return_value = notification_repository_instance
        notification_repository_instance.fetch_last_notification_for_a_validation.return_value = (
            None
        )

        self.mock_validation_response.status = PROBLEM
        self.mock_validation_response.settings = {}
        self.mock_validation_response.details = {
            "traceback": "traceback",
            "use_error_template": True,
        }

        self.mock_settings["notification"] = {"template_error": """*{{traceback}}*"""}

        notify_in_thread(self.mock_validation_response, self.mock_settings)

        mock_implementation.init_thread.assert_called_with(
            "*traceback*", {"notification": {"template_error": "*{{traceback}}*"}}
        )

    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.json")
    @patch("lifeguard.actions.notifications.NotificationRepository")
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_send_notify_in_thread_not_problem_and_not_last_notification(
        self, mock_notification_repository, mock_json
    ):
        notification_repository_instance = MagicMock(name="repository")
        mock_notification_repository.return_value = notification_repository_instance
        notification_repository_instance.fetch_last_notification_for_a_validation.return_value = (
            None
        )

        self.mock_validation_response.status = NORMAL

        notify_in_thread(self.mock_validation_response, self.mock_settings)

        mock_json.dumps.assert_not_called()

    @patch("lifeguard.actions.notifications.deepcopy")
    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.json")
    @patch("lifeguard.actions.notifications.HistoryRepository")
    @patch("lifeguard.actions.notifications.NotificationRepository")
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_send_notify_open_thread(
        self,
        mock_notification_repository,
        mock_history_repository,
        mock_json,
        mock_deepcopy,
    ):
        history_repository_instance = MagicMock(name="history_repository")
        mock_history_repository.return_value = history_repository_instance

        notification_repository_instance = MagicMock(name="repository")
        mock_notification_repository.return_value = notification_repository_instance
        notification_repository_instance.fetch_last_notification_for_a_validation.return_value = (
            None
        )

        mock_json.dumps.return_value = "{}"
        mock_deepcopy.return_value = "{}"

        self.mock_validation_response.validation_name = "validation_name"
        self.mock_validation_response.status = PROBLEM
        self.mock_validation_response.details = {}

        notify_in_thread(
            self.mock_validation_response, {"notification": {"add_to_history": True}}
        )

        mock_deepcopy.assert_called_with("{}")

        mock_implementation.init_thread.assert_called_with(
            "{}", {"notification": {"add_to_history": True}}
        )

        notification_occurrence = (
            history_repository_instance.append_notification.call_args[0][0]
        )

        self.assertEqual(notification_occurrence.notification_type, "init_thread")
        self.assertEqual(notification_occurrence.validation_name, "validation_name")
        self.assertEqual(notification_occurrence.details, {})
        self.assertEqual(notification_occurrence.status, "PROBLEM")

    @patch("lifeguard.actions.notifications.deepcopy")
    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.json")
    @patch("lifeguard.actions.notifications.HistoryRepository")
    @patch("lifeguard.actions.notifications.NotificationRepository")
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_send_notify_close_thread(
        self,
        mock_notification_repository,
        mock_history_repository,
        mock_json,
        mock_deepcopy,
    ):
        history_repository_instance = MagicMock(name="history_repository")
        mock_history_repository.return_value = history_repository_instance

        notification_repository_instance = MagicMock(name="repository")
        mock_notification_repository.return_value = notification_repository_instance
        notification_repository_instance.fetch_last_notification_for_a_validation.return_value = NotificationStatus(
            "validation", {"not implemented": "thread"}
        )

        mock_json.dumps.return_value = "{}"
        mock_deepcopy.return_value = "{}"

        self.mock_validation_response.status = NORMAL
        self.mock_validation_response.details = {}

        notify_in_thread(
            self.mock_validation_response, {"notification": {"add_to_history": True}}
        )

        mock_deepcopy.assert_called_with("{}")

        mock_implementation.close_thread.assert_called_with(
            "thread", "{}", {"notification": {"add_to_history": True}}
        )

        notification_occurrence = (
            history_repository_instance.append_notification.call_args[0][0]
        )

        self.assertEqual(notification_occurrence.notification_type, "close_thread")

    @patch("lifeguard.actions.notifications.deepcopy")
    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.json")
    @patch("lifeguard.actions.notifications.HistoryRepository")
    @patch("lifeguard.actions.notifications.NotificationRepository")
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_send_notify_update_thread(
        self,
        mock_notification_repository,
        mock_history_repository,
        mock_json,
        mock_deepcopy,
    ):
        history_repository_instance = MagicMock(name="history_repository")
        mock_history_repository.return_value = history_repository_instance

        notification_repository_instance = MagicMock(name="repository")
        mock_notification_repository.return_value = notification_repository_instance
        notification_repository_instance.fetch_last_notification_for_a_validation.return_value = NotificationStatus(
            "validation", {"not implemented": "thread"}
        )

        mock_json.dumps.return_value = "{}"
        mock_deepcopy.return_value = "{}"

        self.mock_validation_response.status = PROBLEM
        self.mock_validation_response.details = {}

        notify_in_thread(
            self.mock_validation_response, {"notification": {"add_to_history": True}}
        )

        mock_deepcopy.assert_called_with("{}")

        mock_implementation.update_thread.assert_called_with(
            "thread", "{}", {"notification": {"add_to_history": True}}
        )

        notification_occurrence = (
            history_repository_instance.append_notification.call_args[0][0]
        )

        self.assertEqual(notification_occurrence.notification_type, "update_thread")

    @patch("lifeguard.actions.notifications.deepcopy")
    @patch("lifeguard.actions.notifications.NOTIFICATION_METHODS", NOTIFICATION_METHODS)
    @patch("lifeguard.actions.notifications.json")
    @patch("lifeguard.actions.notifications.NotificationRepository")
    @patch("lifeguard.actions.notifications.logger", mock_logger)
    def test_should_not_update_thread_if_into_interval(
        self, mock_notification_repository, mock_json, mock_deepcopy
    ):
        notification_repository_instance = MagicMock(name="repository")
        mock_notification_repository.return_value = notification_repository_instance
        notification_repository_instance.fetch_last_notification_for_a_validation.return_value = NotificationStatus(
            "validation", {"not implemented": "thread"}
        )

        mock_json.dumps.return_value = "{}"
        mock_deepcopy.return_value = "{}"

        self.mock_validation_response.status = PROBLEM
        self.mock_validation_response.details = {}

        notify_in_thread(
            self.mock_validation_response,
            {"notification": {"update_thread_interval": 60}},
        )

        mock_implementation.update_thread.assert_not_called()



"""
Lifeguard core settings
"""
import re
from os import environ


class AttributeNotFoundInSettings(Exception):
    """Raised when the entry not found in settings"""


class SettingsManager:
    """
    Settings manager
    """

    def __init__(self, settings):
        self.settings = settings

        for entry in self.settings:
            setattr(self, "__{}".format(entry.lower()), self.__get_value(entry))

    def __get_value(self, name):
        options = self.settings[name]
        value = environ.get(name, options["default"])
        if "type" in options:
            return self.__set_type(options, value)
        return value

    def __search_dynamic_attribute(self, name):
        for entry in self.settings:
            if re.match(entry, name):
                return environ.get(name, self.settings[entry]["default"])

        raise AttributeNotFoundInSettings("{} not found".format(name))

    @staticmethod
    def __set_type(setting, value):
        """
        Convert setting value to explict type.
        Types should be passed as string.
        Allowed types: int, float, str, bool, validation_list.

            Parameters:
                    setting (dict): Value of the type to convert
                    value (object): Value to be converted

            Returns:
                    value (object): the value in the specified type when type is allowed
        """
        _format_funcs = {
            "bool": lambda x: str(x).lower() == "true",
            "validation_list": lambda x: [
                validation for validation in x.split(",") if validation
            ],
            "list": lambda line: [entry for entry in line.split(",") if entry],
            "int": int,
            "float": float,
            "str": str,
        }
        if setting["type"] in _format_funcs.keys():
            return _format_funcs[setting["type"]](value)

        return value

    def read_value(self, name):
        """
        Read value from settings
        """
        if hasattr(self, "__{}".format(name.lower())):
            return getattr(self, "__{}".format(name.lower()))

        return self.__search_dynamic_attribute(name)


SETTINGS_MANAGER = SettingsManager(
    {
        "LIFEGUARD_SERVER_PORT": {
            "default": "5567",
            "type": "int",
            "description": "Lifeguard server port number",
        },
        "LIFEGUARD_DIRECTORY": {
            "default": ".",
            "description": "Location of validations and others resources",
        },
        "LIFEGUARD_LOG_LEVEL": {
            "default": "INFO",
            "description": "Sets the Lifeguard's core log level",
        },
        "LIFEGUARD_LOG_FILE": {
            "default": "",
            "description": "Sets lifeguard logs file",
        },
        "LIFEGUARD_LOG_MAX_BYTES": {
            "default": "10000000",
            "type": "int",
            "description": "Rotating log file max bytes",
        },
        "LIFEGUARD_LOG_BACKUP_COUNT": {
            "default": "2",
            "type": "int",
            "description": "Rotating log file backup count",
        },
        "LIFEGUARD_HTTP_PROXY": {
            "default": None,
            "description": "Proxy used to http calls",
        },
        "LIFEGUARD_HTTPS_PROXY": {
            "default": None,
            "description": "Proxy used to https calls",
        },
        "LIFEGUARD_PUBLIC_ADDRESS": {
            "default": "http://localhost:5567",
            "description": "Address to access dashboard",
        },
        "LIFEGUARD_EMAIL_SMTP_USER": {
            "default": "smtp_user",
            "description": "SMTP user used in email action",
        },
        "LIFEGUARD_EMAIL_SMTP_PASSWD": {
            "default": "smtp_passwd",
            "description": "SMTP passwd used in email action",
        },
        "LIFEGUARD_EMAIL_SMTP_SERVER": {
            "default": "smtp_server",
            "description": "SMTP server used in email action",
        },
        "LIFEGUARD_EMAIL_SMTP_PORT": {
            "default": "smtp_port",
            "description": "SMTP port used in email action",
        },
        "LIFEGUARD_RUN_ONLY_VALIDATIONS": {
            "default": "",
            "type": "validation_list",
            "description": "A comma separated list with validations name to be run",
        },
        "LIFEGUARD_SKIP_VALIDATIONS": {
            "default": "",
            "type": "validation_list",
            "description": "A comma separated list with validations name to be skipped",
        },
        "LIFEGUARD_CONTROLLERS_ASSETS_ENABLED": {
            "default": "true",
            "type": "bool",
            "description": "Enable assets controller",
        },
        "LIFEGUARD_SECRET_KEY": {
            "default": "lifeguard_key",
            "description": "Secret key used in session",
        },
        "LIFEGUARD_APPEND_NOTIFICATION_TO_HISTORY": {
            "default": "false",
            "type": "bool",
            "description": "Append all notifications to history",
        },
        "PERMANENT_SESSION_LIFETIME": {
            "default": "2678400",
            "type": "int",
            "description": "Session expiration time",
        },
    }
)

LIFEGUARD_SERVER_PORT = SETTINGS_MANAGER.read_value("LIFEGUARD_SERVER_PORT")
LIFEGUARD_DIRECTORY = SETTINGS_MANAGER.read_value("LIFEGUARD_DIRECTORY")
LIFEGUARD_PUBLIC_ADDRESS = SETTINGS_MANAGER.read_value("LIFEGUARD_PUBLIC_ADDRESS")
LIFEGUARD_EMAIL_SMTP_USER = SETTINGS_MANAGER.read_value("LIFEGUARD_EMAIL_SMTP_USER")
LIFEGUARD_EMAIL_SMTP_PASSWD = SETTINGS_MANAGER.read_value("LIFEGUARD_EMAIL_SMTP_PASSWD")
LIFEGUARD_EMAIL_SMTP_SERVER = SETTINGS_MANAGER.read_value("LIFEGUARD_EMAIL_SMTP_SERVER")
LIFEGUARD_EMAIL_SMTP_PORT = SETTINGS_MANAGER.read_value("LIFEGUARD_EMAIL_SMTP_PORT")

LIFEGUARD_CONTROLLERS_ASSETS_ENABLED = SETTINGS_MANAGER.read_value(
    "LIFEGUARD_CONTROLLERS_ASSETS_ENABLED"
)

LIFEGUARD_SECRET_KEY = SETTINGS_MANAGER.read_value("LIFEGUARD_SECRET_KEY")

LOG_LEVEL = SETTINGS_MANAGER.read_value("LIFEGUARD_LOG_LEVEL")
LOG_FILE = SETTINGS_MANAGER.read_value("LIFEGUARD_LOG_FILE")
LOG_BACKUP_COUNT = SETTINGS_MANAGER.read_value("LIFEGUARD_LOG_BACKUP_COUNT")
LOG_MAX_BYTES = SETTINGS_MANAGER.read_value("LIFEGUARD_LOG_MAX_BYTES")
HTTP_PROXY = SETTINGS_MANAGER.read_value("LIFEGUARD_HTTP_PROXY")
HTTPS_PROXY = SETTINGS_MANAGER.read_value("LIFEGUARD_HTTPS_PROXY")

LIFEGUARD_RUN_ONLY_VALIDATIONS = SETTINGS_MANAGER.read_value(
    "LIFEGUARD_RUN_ONLY_VALIDATIONS"
)
LIFEGUARD_SKIP_VALIDATIONS = SETTINGS_MANAGER.read_value("LIFEGUARD_SKIP_VALIDATIONS")
LIFEGUARD_APPEND_NOTIFICATION_TO_HISTORY = SETTINGS_MANAGER.read_value(
    "LIFEGUARD_APPEND_NOTIFICATION_TO_HISTORY"
)
PERMANENT_SESSION_LIFETIME = SETTINGS_MANAGER.read_value("PERMANENT_SESSION_LIFETIME")



NORMAL = "NORMAL"
WARNING = "WARNING"
PROBLEM = "PROBLEM"

ACTION_STATUSES = [NORMAL, WARNING, PROBLEM]


def change_status(old, new):
    """
    Change status by severity
    """
    if ACTION_STATUSES.index(new) > ACTION_STATUSES.index(old):
        return new
    return old



import unittest
from unittest.mock import MagicMock, patch

from lifeguard.actions.database import save_result_into_database


class TestActionDatabase(unittest.TestCase):
    @patch("lifeguard.actions.database.ValidationRepository")
    def test_save_result_into_database(self, mock_validation_repository):
        mock_validation_repository_instance = MagicMock(name="instance")
        mock_validation_repository.return_value = mock_validation_repository_instance

        save_result_into_database("validation_response", {})

        mock_validation_repository_instance.save_validation_result.assert_called_with(
            "validation_response"
        )



import json
import traceback
from datetime import datetime, timedelta

from flask import Flask, make_response

from lifeguard import NORMAL, change_status
from lifeguard.controllers import custom_controllers, login_required
from lifeguard.logger import lifeguard_logger as logger
from lifeguard.repositories import ValidationRepository
from lifeguard.settings import LIFEGUARD_SECRET_KEY, PERMANENT_SESSION_LIFETIME
from lifeguard.validations import VALIDATIONS, ValidationResponseEncoder

APP = Flask(__name__)
APP.secret_key = LIFEGUARD_SECRET_KEY
APP.config["PERMANENT_SESSION_LIFETIME"] = timedelta(seconds=PERMANENT_SESSION_LIFETIME)


def make_json_response(content):
    """
    Make default json response
    """
    response = make_response(content)
    response.headers["Content-Type"] = "application/json"
    return response


def build_global_status(complete=False):
    """
    Build global status method
    """
    try:
        response = {"status": NORMAL, "validations": []}
        repository = ValidationRepository()
        for validation in repository.fetch_all_validation_results():
            response["status"] = change_status(response["status"], validation.status)
            if complete or validation.status != NORMAL:
                response["validations"].append(validation)
        return make_json_response(ValidationResponseEncoder().encode(response))
    except Exception:
        logger.error(
            "error on execute get status",
            extra={"traceback": traceback.format_exc()},
        )
        return make_json_response(json.dumps({"error": traceback.format_exc()}))


@APP.route("/lifeguard/status", methods=["GET"])
@login_required
def get_status():
    """
    Return global status
    """
    return build_global_status()


@APP.route("/lifeguard/status/complete", methods=["GET"])
@login_required
def get_status_complete():
    """
    Return global status with details
    """
    return build_global_status(True)


@APP.route("/lifeguard/validations/<validation>", methods=["GET"])
@login_required
def get_validation(validation):
    repository = ValidationRepository()
    result = repository.fetch_last_validation_result(validation)
    return make_json_response(ValidationResponseEncoder().encode(result))


@APP.route("/lifeguard/validations/<validation>/execute", methods=["POST"])
@login_required
def execute_validation(validation):
    try:
        result = VALIDATIONS[validation]["ref"]()
        result.last_execution = datetime.now()
        return make_json_response(ValidationResponseEncoder().encode(result))
    except Exception:
        logger.error(
            "error on execute validation %s",
            validation,
            extra={"traceback": traceback.format_exc()},
        )
        return json.dumps({"error": traceback.format_exc()})


APP.register_blueprint(custom_controllers)
